#
#-#####################################################################
#
# This is Version 2 of the Blankety-Blank Application. Version 2 has a single blanking style rather than several as used
# previously. The look and feel has also been streamlined. Also, code is in place for launching this application from the SciTE
# Tools menu but that use will not be employed due to erratic screen blanking that is caused from a bug in Scite Ver 1.4 [2004] 
# which we are currently using.
#
#-#####################################################################
#
# EDITS:
#
# 04/27/16 [htruax]: Changed read_doc_boxes( ) to look for '# ***** Calculations *****' as an indicator to stop
# reading '# ***** Field Data *****' boxes. The '# ***** Field Data *****' section has been moved from just before the
# Profiles to just after them.
#
#-##########################################################
#

require 'tk'

#
#-#####################################################################
#
# For displaying error messages.
#
require 'Win32API'

#
# Button styles:
#
$MB_OK = 0
$MB_OKCANCEL = 1
$MB_ABORTRETRYIGNORE = 2
$MB_YESNOCANCEL = 3
$MB_YESNO = 4
$MB_RETRYCANCEL = 5

#
# Icon styles:
#
$MB_ICONHAND = 16
$MB_ICONQUESTION = 32
$MB_ICONEXCLAMATION = 48
$MB_ICONASTERISK = 64

#
# Default button:
#
$MB_DEFBUTTON1 = 0
$MB_DEFBUTTON2 = 256
$MB_DEFBUTTON3 = 512

#
# Return values:
#
$IDOK = 1
$IDCANCEL = 2
$IDABORT = 3
$IDRETRY = 4
$IDIGNORE = 5
$IDYES = 6
$IDNO = 7

def messageBox(message, title = 'Message', style = 0)
	mb = Win32API.new("user32", "MessageBox", ['L', 'P', 'P', 'L'], 'L')
	return mb.Call(0, message, title, style)
end

#
#-#####################################################################
#
# Window drawing section.
#
root = TkRoot.new() { title "BLANKETY BLANK Ver 2" }

root.geometry('+29+80')

#
#-#####################################################################
#
# List Frame: Frame for the Code boxes to be blanked and the Document boxes.
#
list_frame = TkFrame.new(root).pack(:side=>"top", :pady=>5)

#
#-#####################################################################
#
# List Frame: Code Frame - Boxes from the supplied Ruby Code.
#
code_frame = TkFrame.new(list_frame, :borderwidth=>5, :relief=>:ridge).pack(:side=>"left", :fill=>"x", :padx=>5)

code_label = TkLabel.new(code_frame, :text=>"Ruby Code: Single / Multi Blank Capability", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

#
#-#################################################
#
# Code Frame: Code Boxes
#
code_box_list_frame = TkFrame.new(code_frame, :borderwidth=>3, :relief=>:ridge).pack("side"=>"left", "pady" =>5, "padx"=>5)

code_box_list_label = TkLabel.new(code_box_list_frame, :text=>"Box List", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

code_box_sel_butt_frame = TkFrame.new(code_box_list_frame).pack("side"=>"bottom")

code_box_sel_mode_button = TkButton.new(code_box_sel_butt_frame, :text=>"Select Single", :background=>"lightgrey", :foreground=>"blue", :width=>11, "font"=>"arial 10").pack(:pady=> 5)

code_box_sel_del_frame = TkFrame.new(code_box_sel_butt_frame).pack(:pady=> 5)

code_box_clear_button = TkButton.new(code_box_sel_del_frame, :text=>"Delete List", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"left", :padx=> 5)

code_box_clear_mode_button = TkButton.new(code_box_sel_del_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"right", :padx=>4)

code_box_list_bar_vertical = TkScrollbar.new(code_box_list_frame).pack("side"=>"right", "fill"=>"y")

code_box_list = TkListbox.new(code_box_list_frame, "background"=>"white", "exportselection" => 0, "width"=>15, "height"=>8, :selectmode=>"extended", "font"=>"arial 10").pack(:side=>"left")

code_box_list.yscrollbar(code_box_list_bar_vertical)

#
#-#################################################
#
# Code Frame: Exclusion Boxes
#
code_excl_box_list_frame = TkFrame.new(code_frame, :borderwidth=>3, :relief=>:ridge).pack(:anchor=>"n", :side=>"left", :pady=>5, :padx=>4)

code_excl_button = TkButton.new(code_excl_box_list_frame, :text=>"Exclusion Boxes", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:pady=>5)

code_excl_del_frame = TkFrame.new(code_excl_box_list_frame).pack(:side=>"bottom")

code_excl_clear_sel_button = TkButton.new(code_excl_del_frame, :text=>"Delete Selection", :background=>"lightgrey", :foreground=>"blue", :width=>12, "font"=>"arial 10").pack(:pady=>5)

code_excl_clear_button = TkButton.new(code_excl_del_frame, :text=>"Delete List", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:side=>"left", :padx=>5, :pady=>5)

code_excl_clear_mode_button = TkButton.new(code_excl_del_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"right", :padx=>4, :pady=>5)

code_excl_box_list_bar_vert = TkScrollbar.new(code_excl_box_list_frame).pack("side"=>"right", "fill"=>"y")

code_excl_box_list = TkListbox.new(code_excl_box_list_frame, "background"=>"white", "exportselection" => 0, "width"=>15, "height"=>7, :selectmode=>"extended", "font"=>"arial 10").pack(:side=>"left")

code_excl_box_list.yscrollbar(code_excl_box_list_bar_vert)

#
#-#####################################################################
#
# List Frame: Document Frame - All boxes from the document.
#
doc_frame = TkFrame.new(list_frame, :borderwidth=>5, :relief=>:ridge).pack(:side=>"left", :fill=>"x", :padx=>5)

doc_label = TkLabel.new(doc_frame, :text=>"Document: Single Blank Only", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

#
#-#################################################
#
# Document Frame: Document Boxes
#
doc_box_list_frame = TkFrame.new(doc_frame, :borderwidth=>3, :relief=>:ridge).pack("side"=>"left", "pady" =>5, "padx"=>5)

doc_box_list_label = TkLabel.new(doc_box_list_frame, :text=>"Box List", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

doc_box_sel_butt_frame = TkFrame.new(doc_box_list_frame).pack("side"=>"bottom")

doc_box_sel_mode_button = TkButton.new(doc_box_sel_butt_frame, :text=>"Select Single", :background=>"lightgrey", :foreground=>"blue", :width=>11, "font"=>"arial 10").pack(:pady=> 5)

doc_box_sel_del_frame = TkFrame.new(doc_box_sel_butt_frame).pack(:pady=> 5)

doc_box_clear_button = TkButton.new(doc_box_sel_del_frame, :text=>"Delete List", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"left", :padx=> 5)

doc_box_clear_mode_button = TkButton.new(doc_box_sel_del_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"right", :padx=>4)

doc_box_list_bar_vertical = TkScrollbar.new(doc_box_list_frame).pack("side"=>"right", "fill"=>"y")

doc_box_list = TkListbox.new(doc_box_list_frame, "background"=>"white", "exportselection" => 0, "width"=>15, "height"=>8, :selectmode=>"extended", "font"=>"arial 10").pack(:side=>"left")

doc_box_list.yscrollbar(doc_box_list_bar_vertical)

#
#-#################################################
#
# Document Frame: Inclusion Boxes
#
doc_incl_box_list_frame = TkFrame.new(doc_frame, :borderwidth=>3, :relief=>:ridge).pack(:anchor=>"n", :side=>"left", :pady=>5, :padx=>4)

doc_incl_button = TkButton.new(doc_incl_box_list_frame, :text=>"Inclusion Boxes", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:pady=>5)

doc_incl_del_frame = TkFrame.new(doc_incl_box_list_frame).pack(:side=>"bottom")

doc_incl_clear_sel_button = TkButton.new(doc_incl_del_frame, :text=>"Delete Selection", :background=>"lightgrey", :foreground=>"blue", :width=>12, "font"=>"arial 10").pack(:pady=>5)

doc_incl_clear_button = TkButton.new(doc_incl_del_frame, :text=>"Delete List", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:side=>"left", :padx=>5, :pady=>5)

doc_incl_clear_mode_button = TkButton.new(doc_incl_del_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"right", :padx=>4, :pady=>5)

doc_incl_box_list_bar_vert = TkScrollbar.new(doc_incl_box_list_frame).pack("side"=>"right", "fill"=>"y")

doc_incl_box_list = TkListbox.new(doc_incl_box_list_frame, "background"=>"white", "exportselection" => 0, "width"=>15, "height"=>7, :selectmode=>"extended", "font"=>"arial 10").pack(:side=>"left")

doc_incl_box_list.yscrollbar(doc_incl_box_list_bar_vert)

#
#-#####################################################################
#
# This frame contains the Multi Blank group size and Custom Blanking areas.
#
misc_frame = TkFrame.new(list_frame, :borderwidth=>5, :relief=>:ridge).pack(:side=>"left", :fill=>"x", :padx=>5)

#
#-#################################################
#
# Custom Blanking:
#
custom_frame = TkFrame.new(misc_frame, :borderwidth=>3, :relief=>:ridge).pack(:side=>"top", :fill=>"x", :pady=>5, :padx=>5)

custom_label = TkLabel.new(custom_frame, :text=>"Custom BLANKETY BLANK Style - Defaults to: if ( )", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>"x")

custom_button_frame = TkFrame.new(custom_frame).pack(:side=>"bottom", :pady=>5, :padx=>5)

cust_clear_button = TkButton.new(custom_button_frame, :text=>"Delete Custom", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:side=>"left", :padx=>5)

cust_clear_mode_button = TkButton.new(custom_button_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"left", :padx=>5)

custom_bar_horiz = TkScrollbar.new(custom_frame, "orient"=>"hor").pack("side"=>"bottom", "fill"=>"x")

custom_bar_text = TkText.new(custom_frame, "wrap"=>"none", "width"=>45, "height"=>1, "font"=>"arial 10").pack("fill"=>"x", "expand"=>true)
custom_bar_text.xscrollbar(custom_bar_horiz)

#
#-#################################################
#
# Multi Blanking:
#
multi_frame = TkFrame.new(misc_frame, :borderwidth=>3, :relief=>:ridge).pack(:side=>"bottom", :fill=>:x, :pady=>5, :padx=>5)

multi_label = TkLabel.new(multi_frame, :text=>"Multi Blank", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>"x")

group_size_frame = TkFrame.new(multi_frame, :borderwidth=>1, :relief=>:ridge).pack(:side=>:left, :pady=>5, :padx=>10)

group_label = TkLabel.new(group_size_frame, :text=>"Group Size", :width=>10, "font"=>"arial 10").pack(:side=>:left)

multi_boxes_group_num = TkVariable.new

multi_boxes_group_num_entry = TkEntry.new(group_size_frame, :textvariable=>multi_boxes_group_num, "width"=>3, "font"=>"arial 10").pack(:side=>:left)

multi_button_frame = TkFrame.new(multi_frame).pack(:side=>:left, :pady=>5, :padx=>5)

multi_clear_button = TkButton.new(multi_button_frame, :text=>"Delete Size", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:side=>"left")

multi_clear_mode_button = TkButton.new(multi_button_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", :font=>"arial 10").pack(:side=>"left", :padx=>10)

#
#-#####################################################################
#
# Main buttons:
#
blank_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"x")

s_blank_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "padx"=>10, "pady"=>5)
s_blank_button = TkButton.new(s_blank_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Single Blank", :foreground=>"blue", "font"=>"arial 10").pack()

m_blank_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "pady"=>5)
m_blank_button = TkButton.new(m_blank_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Multi Blank", :foreground=>"blue", "font"=>"arial 10").pack()

c_blank_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "pady"=>5, "padx"=>10)
c_blank_button = TkButton.new(c_blank_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Custom Blank", :foreground=>"blue", "font"=>"arial 10").pack()

r_blank_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "pady"=>5)
r_blank_button = TkButton.new(r_blank_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Reverse Blank", :foreground=>"blue", "font"=>"arial 10").pack()

clear_input_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "pady"=>5, "padx"=>10)
clear_input_button = TkButton.new(clear_input_button_frame, "width"=>12, :background=>"lightgrey", :text=> "Delete Code", :foreground=>"blue", "font"=>"arial 10").pack()

get_boxes_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "pady"=>5)
get_boxes_button = TkButton.new(get_boxes_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Code Box List", :foreground=>"blue", "font"=>"arial 10").pack()

delete_box_list_button_frame = TkFrame.new(blank_frame).pack("side"=>"left", "pady"=>5, "padx"=>10)
delete_box_list_button = TkButton.new(delete_box_list_button_frame, "width"=>12, :background=>"lightgrey", :text=>"Delete All Lists", :foreground=>"blue", "font"=>"arial 10").pack()

input_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"both", "expand"=>true)

in_text_frame = TkFrame.new(input_frame).pack("side"=>"left", "fill"=>"both", "padx"=>10, "expand"=>true)
in_text_bar_vert = TkScrollbar.new(in_text_frame).pack("side"=>"right", "fill"=>"y")
in_text = TkText.new(in_text_frame, "width"=>113, "height"=>12, "font"=>"arial 10").pack("side"=>"bottom", "fill"=>"both", "expand"=>true)

in_text.yscrollbar(in_text_bar_vert)

in_text.focus

output_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"both", "expand"=>true)

output_label_frame = TkFrame.new(output_frame).pack("side"=>"top")
output_label = TkLabel.new(output_label_frame, :text=>"Code Converted To BLANKETY BLANK Form - NOTE: ' Ctrl-/ ' selects all text.", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack()

out_text_frame = TkFrame.new(output_frame).pack("side"=>"left", "padx" => 10, "fill"=>"both", "expand"=>true)
out_text_bar_vert = TkScrollbar.new(out_text_frame).pack("side"=>"right", "fill"=>"y")
out_text = TkText.new(out_text_frame, "width"=>113, "height"=>12, "font"=>"arial 10").pack("side"=>"bottom", "fill"=>"both", "expand"=>true)

out_text.yscrollbar(out_text_bar_vert)

bottom_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"x", "pady" => 5)

#
# ###############################################################
#
# Globals:
#
cur_dir = "C:/temp/"

box_array = Array.new

# This tells us if this app was started from the SciTE Tools menu. If 'true' we print [output] the blanked/reverse
# blanked code which is read by SciTE and results in the selected Ruby Code being overwritten or if no selected text,
# the output is inserted into the Ruby Code file at the current cursor location. 'false' means this app was started
# manually, so the user needs to copy the blanked/reverse blanked code and paste into the Ruby Code file.
from_SciTE = false

#
# ###############################################################
#
# Methods:
#
def process_errors(error, var, type)

	e_type = (type.eql?("W") ? "Warning" : "Error")

	err_msg = [
	"  #{e_type.upcase} - Duplicate Box: #{var}\t\n\n  This box is already in the list.\n",
	"  #{e_type.upcase} - Duplicate Box: #{var}\t\n\n  This box is in the Exclusion Boxes list.\n",
	"  #{e_type.upcase} - Duplicate Box: #{var}\t\n\n  This box is in the Inclusion Boxes list.\n",
	"  #{e_type.upcase} - #{var} Cannot be a trigger box - This is a calculated box.\n",
	"  Application was started from a SciTE buffer that did not contain the Ruby code from a document.\n"
	]

	#
	# Using bitwise operations, error is set on a bit by bit basis, where each bit represents an error message.
	#
	if (error > 0)

		mask = 1

		#
		# Math.frexp(error) returns a two element array containing the normalized fraction (a Float) and exponent
		# (a Fixnum) of error. In our case, the fraction is 0.5 and the exponent is a power of 2, as our error is 1 x 2
		# raised to 0,1,2,3,... So this returns 0.5 x 2[a power]. So, say error = 4, which is 1 x 2[2], which is the 3rd
		# error in our list. .frexp returns [0.5, 3] - we want the 3 [3rd error], we use 3 - 1 since we are starting
		# from 0 not 1. Just doing this so that I do not have to keep updating the 0.upto( ) each time I add a new error.
		#
		0.upto(Math.frexp(error)[1]-1) do |shift|
			if (((mask << shift) & error) > 0)
				messageBox(err_msg[shift], "#{e_type}!", $MB_ICONEXCLAMATION|$MB_OK)
			end
		end
	end
end

def get_file_name(dir, type)

	ftypes = case (type)
			when ("RBW"): [ ["RBW files", '.rbw'] ]
		end

	Tk.getOpenFile(:filetypes=>ftypes, :initialdir=>dir)
end

def read_doc_boxes(source)

	box_array = Array.new
	tmp_array = Array.new

	File.open(source, "r") do |src_file|

		#while ((line = src_file.gets) and (line !~ /\*+\s*Profile Classes\s*\*+/))
		while ((line = src_file.gets) and (line !~ /\*+\s*Calculations\s*\*+/))
			box_array.push($1) if (line =~ /(box\d+_\d+)/)
		end
	end

	#
	# Sort the array:
	#
	box_array.each { |box|
		box.scan(/box(\d+)_(\d+)/) { tmp_array.push([$1.to_i, $2.to_i]) }
	}

	tmp_array.sort!
	box_array.clear

	tmp_array.each_index { |index|
		box_array.push("box#{tmp_array[index][0]}_#{tmp_array[index][1]}")
	}

	box_array
end

#
# This will reverse the blanked code to the original form. 05/13/14 [HET]: Keeping this method pretty much the same regarding
# looking for the various blanking patterns to reverse [April/Bob/Harry] as these have been in use for several years.
#
def reverse_blank_button(from_SciTE, out_text, in_text)

	box_array = Array.new
	code_array = Array.new
	input_array = Array.new
	temp_array = Array.new
	print_array = Array.new
	blank_array = Array.new
	comment_array = Array.new

	output_str = String.new

	# Clear the output area.
	#~ out_text.delete("1.0", "end")
	#~ out_text.delete("1.0", "end") unless (from_SciTE)

	#
	# input_array is set via splitting the supplied code at the newline character, each array element is a line of data.
	#
	input_array = in_text.get("1.0", "end").split(/\n/)

	return if (input_array.length == 0)

	#
	# Process the input_array by looking for a blankety-blank pattern, or patterns in the case of a multi-blank situation.
	#
	process_code = 0
	reverse_code = 0
	end_cnt = 0

	process_print = false
	process_blank = false
	check_for_end_exp = false

	pattern_string = output_str = ""

	input_array.each { |code_line|
#~ out_text.insert("end", "At top - code_line = #{code_line}\n")
#~ out_text.insert("end", "At top - pattern_string = #{pattern_string}\n")

		#
		# Look for the blankety-blank pattern(s) - April/Bob/Harry/Custom.
		#
		# An 'if' may or may not be the start of a blankety blank pattern, we start by assuming it is, but we also keep
		# a copy of each code line in case it turns out not to be a blankety blank pattern, in which case we will turn off the
		# pattern recognition flag and output the previously copied code as it was inputted.
		#
		# For blankety blank patterns in process:
		#
		# We will skip blank lines (new line char is stripped, so a blank line will have 0 length in the input_array), we will
		# oversee the line spacing when we output the reversed code.
		#
		if (!pattern_string.empty?)
			temp_array.push(code_line)
			next if (code_line.strip.length == 0)
		end

		case code_line.strip

			when (/\A\s*if(\s+|\()/)

				#
				# NOTE: An 'if' statement - could have nothing between the parens (empty custom blanking).
				# ALSO: This could be the start of a multi-blankety blank component, or a stand-alone blankety blank
				# pattern.
				#
				if (pattern_string.empty?)

					temp_array.push(code_line)
					pattern_string = "if"

				elsif ((pattern_string == "if blank else") or (pattern_string == "if"))

					#
					# This 'if' is in the code section, it is not a blankety blank 'if'.
					#
					end_cnt += 1

					#
					# First get any comments since we saw the 'initial if'.
					#
					if (pattern_string == "if")
						comment_array.each { |comment|
							code_array.push(comment)
						}
					end

					code_array.push(code_line.sub(/\t/,''))
					pattern_string += " code"

				elsif ((pattern_string == "if blank else code") or (pattern_string == "if code"))
					end_cnt += 1
					code_array.push(code_line.sub(/\t/,''))
				else
					#
					# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
					#
					process_code = 1
				end

			when (/\A\s*else\z/)

				#
				# We need to be careful here, as an 'else' could go with a code section 'if', or be the 'else'
				# that goes with the blankety blank format.
				#
				if (pattern_string == "if blank")
					pattern_string += " else" 
				elsif (pattern_string == "if code")
					if (end_cnt > 0)
						code_array.push(code_line.sub(/\t/,''))
					else
						pattern_string += " else" 
					end
				elsif (pattern_string == "if blank else code")
					if (end_cnt > 0)
						code_array.push(code_line.sub(/\t/,''))
					else
						#
						# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
						#
						process_code = 1
					end
				else
					#
					# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
					#
					process_code = 1
				end

			when (/\A\s*end\z/)

				#
				# We need to be careful here, as an 'end' could go with a code section 'if', or be the 'end'
				# that goes with the blankety blank format, or is an 'end' to a 'case', 'while', 'until', 'for',
				# 'unless', or 'loop do' expression (yikes !).
				#
				if (pattern_string == "if code else blank")
					pattern_string += " end" 
				elsif (pattern_string == "if blank else code")
					if (end_cnt > 0)
						code_array.push(code_line.sub(/\t/,''))
						end_cnt -= 1
					else
						pattern_string += " end" 
					end
				elsif (pattern_string == "if code")
					if (end_cnt > 0)
						code_array.push(code_line.sub(/\t/,''))
						end_cnt -= 1
					else
						#
						# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
						#
						process_code = 1
					end
				else
					#
					# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
					#
					process_code = 1
				end

			when (/(\A\s*print\s*')(box)(\d+)(_)(\d+)(',\s*box\d+_\d+\z)/)
#~ out_text.insert("end", "a print: $3 = #{$3}, $5 = #{$5}\n")
#~ out_text.insert("end", "a print: pattern_string = #{pattern_string}\n")

				#
				# The 'print' section comes at the very end of a properly formed blankety blank formation.
				# If this is a blankety blank print, check to see if this is the end of the pattern by comparing to the
				# blanked boxes array.
				#
				if ((pattern_string == "if blank else code end") or (pattern_string == "if code else blank end"))
					pattern_string += " print"
					process_print = true

				elsif ((pattern_string == "if blank else code end print") or (pattern_string == "if code else blank end print"))

					process_print = true

				else
					#
					# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
					#
					process_code = 1
				end

				if (process_print)

					#
					# This print statement is a blankety blank print (so far), check to see if we are done with
					# the pattern.
					#
					print_array.push([$3.to_i,$5.to_i])
					print_array.sort!
#~ out_text.insert("end", "print_array.push([#{$3},#{$5}])\n")

					if (blank_array.eql?(print_array))

						#
						# Done with the current blankety blank pattern.
						#
						process_code = 2
					end

					process_print = false
				end

			when (/(\A\s*)(box)(\d+)(_)(\d+)(\s*=\s*''\z)/)

				#
				# A box is blanked - it is conceivable that this could be a valid code section component.
				# If the very first line in the blankety blank code section is a blank assignment to a box,
				# this will be construed as the start of the blankety blank blank section, so to get around
				# this 'mode switch' the user can make this line a comment and it will be properly treated
				# as a code section component.
				#
				if ((pattern_string == "if") or (pattern_string == "if code else"))
					pattern_string += " blank"
					process_blank = true
				elsif ((pattern_string == "if blank") or (pattern_string == "if code else blank"))
					process_blank = true
				elsif ((pattern_string == "if code") or (pattern_string == "if blank else code"))
					code_array.push(code_line.sub(/\t/,''))
				else
					#
					# Not a blankety blank pattern - output the temp array and reset the pattern string to empty.
					#
					process_code = 1
				end

				if (process_blank)
					blank_array.push([$3.to_i,$5.to_i])
					blank_array.sort!
					process_blank = false
				end

			when (/\A\s*#/)

				#
				# A comment - ignore comments in blankety-blank blank and print sections.
				#
				case pattern_string

					when /\Aif\z/

						#
						# Remember the comment, use it if we move into a code section - we do not yet know
						# for sure.
						#
						comment_array.push(code_line.sub(/\t/,''))

					when /\Aif blank\z|\Aif code else\z|\Aif code else blank\z|\Aif blank else code end\z|\Aif blank else code end print\z|\Aif code else blank end\z|\Aif code else blank end print\z/

						#
						# Ignore the comment.
						#

					when /\Aif code\z|\Aif blank else\z|\Aif blank else code\z/

						#
						# Process the comment - switch to code mode if appropriate.
						#
						code_array.push(code_line.sub(/\t/,''))
						pattern_string += " code" if (pattern_string == "if blank else")

					else
						#
						# Not a blankety blank pattern - output the temp array and reset the pattern string
						# to empty.
						#
						process_code = 1
				end

			else

				#
				# Something other than an 'if/else/end/print/blank/comment' - defaults to a code section component
				# if we are processing a blankety blank component, else, we just output whatever it is and go to the
				# next index.
				#
				case pattern_string

					when /\Aif code\z|\Aif blank else code\z/

						code_array.push(code_line.sub(/\t/,''))

						#
						# We just inserted into the code_array, check if we inserted an expression that also
						# uses an 'end' keyword.
						#
						check_for_end_exp = true

					when /\Aif\z|\Aif blank else\z/

						#
						# First get any comments since we saw the 'initial if'.
						#
						if (pattern_string == "if")
							comment_array.each { |comment|
								code_array.push(comment)
							}
						end

						code_array.push(code_line.sub(/\t/,''))

						#
						# We just inserted into the code_array, check if we inserted an expression that also
						# uses an 'end' keyword.
						#
						check_for_end_exp = true

						pattern_string += " code"
#~ out_text.insert("end", "else:   code_line = #{code_line}\n")
#~ out_text.insert("end", "        pattern_string = #{pattern_string}\n")

					else
						#
						# Not a blankety blank pattern - output the temp array and reset the pattern
						# string to empty.
						#
						process_code = 1
				end

				if (check_for_end_exp)

					if (code_line =~ /\A\s*for\s+|\A\s*unless(\s+|\()|\A\s*case(\s+|\()|\A\s*while(\s+|\()|\A\s*until(\s+|\()|\A\s*loop\s+do/)
						end_cnt += 1
					end

					check_for_end_exp = false
				end
		end

		case process_code

			when 1

				#
				# Blankety blank pattern is not "complete", output the temp array if it contains anything, output the
				# current code_line value if we are not working on a pattern (pattern_string being empty, thus code_line
				# value will not be stored in temp_array).
				#
				temp_array.each { |temp_code|

					#~ out_text.insert("end", "#{temp_code}\n")
					output_str += "#{temp_code}\n"
				}

				#~ out_text.insert("end", "#{code_line}\n") if (pattern_string.empty?)
				output_str += "#{code_line}\n" if (pattern_string.empty?)

			when 2
#~ out_text.insert("end", "process_code = #{process_code}\n")

				#
				# We found a complete pattern - Now use the code array to produce the blankety blank reversal output.
				#
				end_cnt = 0
				prev_box = ''

				0.upto(code_array.length - 1) { |index|
			
					#~ out_text.insert("end", "#{code_array[index]}\n")
					output_str += "#{code_array[index]}\n"

					case code_array[index]

						when (/(\A\s*)(box)(\d+)(_)(\d+)(\s*(\+?|-?)=.*\z)/)

							#
							# Box assignment - build a list of boxes if we are inside an expression structure that terminates
							# with an 'end' keyword - for example: 'if'/'unless'/'case'/'while'/'until'/'for'/'loop do'.
							#
							box_array.push([$3.to_i,$5.to_i]) if ((!box_array.include?([$3.to_i,$5.to_i])) and (end_cnt > 0))

							this_box = $2+$3+$4+$5

							if (index == code_array.length - 1)

								#~ out_text.insert("end", "print '#{this_box}', #{this_box}\n")
								output_str += "print '#{this_box}', #{this_box}\n"

							elsif (code_array[index + 1] =~ /(\A\s*)(box\d+_\d+)(\s*(\+?|-?)=.*\z)/)
								if (end_cnt == 0)
									if (this_box != $2)
										#~ out_text.insert("end", "print '#{this_box}', #{this_box}\n\n")
										output_str += "print '#{this_box}', #{this_box}\n\n"
									end
								end
							else
								#~ out_text.insert("end", "print '#{this_box}', #{this_box}\n") if (end_cnt == 0)
								output_str += "print '#{this_box}', #{this_box}\n"  if (end_cnt == 0)
								#~ out_text.insert("end", "\n")
								output_str += "\n"
							end

						when (/\A\s*if(\s+|\()|\A\s*for\s+|\A\s*unless(\s+|\()|\A\s*case(\s+|\()|\A\s*while(\s+|\()|\A\s*until(\s+|\()|\A\s*loop\s+do/)

							#
							# An 'if'/'for'/etc.. statement - print boxes at end of outer 'if'/'for'/etc.. statement
							# (may be nesting). A list of all boxes that are assigned to will be built and then used to
							# produce the print statements once the outer 'end' statement is encountered (end_cnt == 0).
							#
							end_cnt += 1
							#~ out_text.insert("end", "\n")
							output_str += "\n"

						when (/\A\s*end\z/)

							#
							# Decrement the end_cnt and if equal to zero, output the print box statements.
							#
							end_cnt -= 1 if (end_cnt > 0)

							if (end_cnt == 0)

								#~ out_text.insert("end", "\n")
								output_str += "\n"

								box_array.sort!

								box_array.each_index { |i|

									box = "box#{box_array[i][0]}_#{box_array[i][1]}"
									#~ out_text.insert("end", "print '#{box}', #{box}\n")
									output_str += "print '#{box}', #{box}\n"
								}

								#
								# Empty the box_array.
								#
								box_array.clear
							end

							#~ out_text.insert("end", "\n") if (index != code_array.length - 1)
							output_str += "\n" if (index != code_array.length - 1)
						else
							#~ out_text.insert("end", "\n")
							output_str += "\n"
					end
				}
		end
#~ out_text.insert("end", "At bottom - process_code = #{process_code}\n")
#~ out_text.insert("end", "At bottom - pattern_string = #{pattern_string}\n")

		if (process_code > 0)

			#
			# Either we determined that the 'pattern' we were working on turned out not to be a blankety blank
			# pattern, or we just processed a blankety blank pattern. Init to recognise and process the next pattern.
			#
			pattern_string = ''
			temp_array.clear
			blank_array.clear
			print_array.clear
			code_array.clear
			comment_array.clear
			process_code = 0
		end
	}

	#
	# There could possibly be some leftover code in the temp_array.
	#
	temp_array.each { |temp_code|

		#~ out_text.insert("end", "#{temp_code}\n")
		output_str += "#{temp_code}\n"
	}

	if (from_SciTE)

		# Program output is captured by SciTE and will overwrite any selected text or will be inserted at the cursor position
		# in the Ruby Code file.
		print output_str.delete("\r")
		#~ out_text.get("1.0", "end").split(/\n/).each do |line| print "#{line}\n" end
	else
		out_text.delete("1.0", "end")
		out_text.insert("end", output_str)

		# Set the focus to the output text window, and put the cursor at the first char on line one.
		out_text.focus
		out_text.mark_set('insert', '1.0')
	end
end

def blank_button(from_SciTE, mode, out_text, in_text, code_excl_box_list, doc_incl_box_list, custom_bar_text, group_num)

	blank_array = Array.new
	blank_page_num_array = Array.new
	print_array = Array.new
	code_array = Array.new
	incl_box_array = Array.new

	input_str = String.new

	box = String.new

	error_code = 0
	group_cnt = 1

	# Clear the output area, and check if we can proceed.
	out_text.delete("1.0", "end") unless (from_SciTE)

	# Just catching a user goof.
	group_num = 0 if (group_num < 0)

	# Multi-blanking and Custom blanking ignores the incl_box_array.
	incl_box_array = doc_incl_box_list.get(0, "end") if (mode == "single")

	input_str = in_text.get("1.0", "end").strip

	if (input_str.strip.length > 0)

		input_str.each { |code_line|

			case
				when (code_line =~ /(\s*print +')(box)(\d+)(_)(\d+)(', +box\d+_\d+)/)

					blank_page_num_array.push([$3.to_i,$5.to_i]) unless (blank_page_num_array.include?([$3.to_i, $5.to_i]))
				when (!code_line.strip.empty?): code_array.push(code_line)

				when (mode == "multiple")

					# We are multi-blanking, we need to determine if it is time to do it now, as blank lines are significant.
					if ((group_num == 0) or (group_num == group_cnt))

						# First produce a sorted blank_array and use it to make the print_array - should have done this
						# differently, but time is short and I am tired and it is 'Friday'.
						blank_page_num_array.sort!

						blank_page_num_array.each_index { |i|

							box = "box#{blank_page_num_array[i][0]}_#{blank_page_num_array[i][1]}"
							blank_array.push(box)
							print_array.push("print '#{box}', #{box}")
						}
	
						do_the_blank_thing(from_SciTE, mode, out_text, code_excl_box_list, doc_incl_box_list, custom_bar_text, print_array, blank_array, code_array)

						# Init for next group.
						group_cnt = 1
						print_array.clear
						blank_array.clear
						blank_page_num_array.clear
						code_array.clear
						out_text.insert("end", "\n")
					else
						group_cnt += 1
					end
			end
		}

		# Check for any blanking that needs to be done, if we are single blanking, we will need to do it now.
		if ((print_array.length + blank_array.length + code_array.length) > 0)

			# First produce a sorted blank_array and use it to make the print_array - should have done this
			# differently, but time is short and I am tired and it is 'Friday'.
			blank_page_num_array.sort!

			blank_page_num_array.each_index { |index|

				box = "box#{blank_page_num_array[index][0]}_#{blank_page_num_array[index][1]}"
				blank_array.push(box)
				print_array.push("print '#{box}', #{box}")
			}

			do_the_blank_thing(from_SciTE, mode, out_text, code_excl_box_list, doc_incl_box_list, custom_bar_text, print_array, blank_array, code_array)
		end

		unless (from_SciTE)

			# Set the focus to the output text window, and put the cursor at the first char on line one.
			out_text.focus
			out_text.mark_set('insert', '1.0')
		end
	end
end

def do_the_blank_thing(from_SciTE, mode, out_text, code_excl_box_list, doc_incl_box_list, custom_bar_text, print_array, blank_array, code_array)

	exclusion_box_array = Array.new
	inclusion_box_array = Array.new
	page_num_array = Array.new

	box = String.new
	box_str = String.new
	output_str = String.new

	# box_str will be the string of boxes for the if ( ) statement.
	box_str = ''

	if (mode == "custom")

		box_str = (custom_bar_text.strip.empty? ? "if ( )" : custom_bar_text)
	else
		#
		# Not doing custom blanking, so build the box_str for the if expression from the ruby code
		# and if not multi-blanking, add the inclusion boxes.
		#
		exclusion_box_array = code_excl_box_list.get(0, "end")

		code_array.each { |code_line|
			code_line.scan(/(box)(\d+)(_)(\d+)/) {
			
			if (!(blank_array.include?($1+$2+$3+$4) or exclusion_box_array.include?($1+$2+$3+$4) or page_num_array.include?([$2.to_i,$4.to_i])))
				page_num_array.push([$2.to_i,$4.to_i])
			end
			}
		}

		unless (mode == "multiple")

			inclusion_box_array = doc_incl_box_list.get(0, "end")

			inclusion_box_array.each { |box|

				box.scan(/(box)(\d+)(_)(\d+)/) {

					#
					# We have already made sure that the Exclusion and Inclusion lists do not intersect.
					#
					case
						when (blank_array.include?($1+$2+$3+$4)): process_errors(8, box, "W")
						when (page_num_array.include?([$2.to_i,$4.to_i])): # do nothing.
						else page_num_array.push([$2.to_i,$4.to_i])
					end
				}
			}
		end

		# Now sort the two-dimensional page_num_array, which will put the boxes in order correctly, note that this
		# is not a string sort, but an integer sort based on page number and box number for each box. A string sort
		# would say that 'box1_1' is greater than 'box10_1'.
		page_num_array.sort!

		# Now put the box parts back together and assign to box_str. box_str now holds all boxes that are not in
		# the exclusion array, not in the blank array, and are not duplicates - they came from the range of boxes
		# encapsulated in the code that is to be blanked, plus any boxes added per user request via the inclusion list
		# [non multi blank only].
		page_num_array.each_index { |i| 
			box_str += "box#{page_num_array[i][0]}_#{page_num_array[i][1]}"
		}

		box_str.gsub!(/(\d)(b)/, '\1 + \2')
	end


	# Now output the blanked code in the desired blanking style. Output will be either to the out_text region of the tool or
	# to standard output via print.
	output_str = ""
	output_str += (mode == "custom" ? "#{box_str}\n" : "if ((#{box_str}).strip.length > 0)\n\n")

	code_array.each { |print_line| output_str += "\t#{print_line}" }
	output_str += "else\n"
	blank_array.each { |print_line| output_str += "\t#{print_line} = ''\n" }
	output_str += "end\n\n"
	print_array.each { |print_line| output_str += "#{print_line}\n" }

	if (from_SciTE)

		# Program output is captured by SciTE and will overwrite any selected text or will be inserted at the cursor position
		# in the Ruby Code file.
		print output_str.delete("\r")
		#~ print output_str
		#~ p output_str
	else
		out_text.insert("end", output_str)
	end
end

def process_doc_file(fname, doc_box_list_label, doc_box_list)

	form_name = fname

	doc_box_list_label.configure("text", "Box List #{File.basename(fname, ".*").upcase}")

	doc_box_list.delete(0, "end")

	tmp_array = read_doc_boxes(fname)

	tmp_array.each { |box| doc_box_list.insert("end", box) }

	# Select (show highlighting) first list element.
	doc_box_list.selection_set(0)
end

def process_input_code(code_box_list, in_text)

	page_num_array = Array.new
	box = String.new
	input_str = String.new

	# Clear the Box List.
	code_box_list.delete(0, code_box_list.index("end"))

	# Get input code.
	input_str = in_text.get("1.0", "end")

	input_str.each { |code_line|
		code_line.scan(/(box)(\d+)(_)(\d+)/) {
			page_num_array.push([$2.to_i,$4.to_i]) unless page_num_array.include?([$2.to_i,$4.to_i])
		}
	}

	# Now sort the two-dimensional page_num_array.
	page_num_array.sort!

	# Now put the box parts back together and insert into the Box List widget.
	page_num_array.each_index { |index|
		code_box_list.insert("end", "box#{page_num_array[index][0]}_#{page_num_array[index][1]}")
	}

	# Select (show highlighting) first list element.
	code_box_list.selection_set(0)
end

#
# This is called when the user either clicks on the 'X' button or selects the Exit menu option to close the application window for
# EITHER a stand alone startup or a startup from the Tools Menu in SciTE.
#
#-###############################
#
# When the app is started from the SciTE Tools Menu:
#
# We use this routine to properly handle app exits that are not initiated from the blank or reverse blank procedures because we
# do not want to overwrite/write to the SciTE file from which this app was started via the SciTE Tools menu command. The app
# also exits [with a value of 0] when the user selects either the Single/Multiple/Reverse blank buttons in which case we DO
# want to overwrite any selected text.
#
# exit(0) tells SciTE that the app closed in a 'good' way so any selected text from the Ruby Code file can be overwritten by the
# app output via the print statement.
#
# exit(1) tells SciTE that the app closed from the Exit Menu or the 'X' button when started from SciTE - we do NOT want
# to overwrite any selected text in the associated Ruby Code file.
#
# The SciTEGlobal.properties file contains the command line "command.replace.selection.6.*=2" which tells SciTE not to overwrite
# selected text when we exit with an argument of 1. The 6 is the command number for this app.
# 
#-###############################
#
# exit(0) is called directly from the button procedures for Single/Multi/Reverse buttons, we do not use this routine for those.
# Var from_SciTE is set when we detect [upon startup] ARGV[0] - set by SciTE when we use the Tools menu. See SciTEGlobal.properties
# for this command [#6]. When this app is started as a stand alone app [not from the Tools menu], from_SciTE is set to 'false'.
#
def exit_app(from_SciTE)

	exit(from_SciTE ? 1 : 0)
end

#
# ###############################################################
#
# Menu:
#
#		Read:
#			Read Ruby Code file to create form box list [displaying box numbers].
#
#		Exit:
#			Exit application.
#
main_menu = TkMenu.new(root)

read_menu = TkMenu.new(main_menu, :tearoff=>false)

#
# This reads the .RBW file and builds an array of Boxes defined in the document. 
#
# The form_box_list is then poulated with the box numbers.
#
read_menu.add('command', :label=> "RBW (Ruby Code) File", :command=> proc {

	tmp_array = Array.new

	fname = get_file_name(cur_dir, "RBW")

	unless (fname.empty?)

		cur_dir = File.dirname(fname)
		process_doc_file(fname, doc_box_list_label, doc_box_list)
	end
})

main_menu.add('cascade', :menu=>read_menu, :label=>"Read")

main_menu.add('command', :label=> "Exit", :command=> proc { exit_app(from_SciTE) })

#
# Adding main_menu to the root window.
#
root.menu(main_menu)

#
# ###############################################################
#
# Button bindings:
#
code_excl_button.bind("Button-1") {

	box_indices_array = Array.new
	exclusion_box_array = Array.new
	inclusion_box_array = Array.new
	tmp_array = Array.new

	exclusion_box_array = code_excl_box_list.get(0, "end")
	inclusion_box_array = doc_incl_box_list.get(0, "end")

	box_indices_array = code_box_list.curselection

	box_indices_array.each { |i|

		box = code_box_list.get(i)

		case
			when (exclusion_box_array.include?(box)): process_errors(1, box, "E")
			when (inclusion_box_array.include?(box)): process_errors(4, box, "E")
			else exclusion_box_array.push(box)
		end
	}

	exclusion_box_array.each { |box|
		box.scan(/(box)(\d+)(_)(\d+)/) { tmp_array.push([$2.to_i,$4.to_i]) }
	}

	tmp_array.sort!
	code_excl_box_list.delete(0, "end")

	tmp_array.each_index { |i|
		code_excl_box_list.insert("end", "box#{tmp_array[i][0]}_#{tmp_array[i][1]}")
	}
}

doc_incl_button.bind("Button-1") {

	box_indices_array = Array.new
	exclusion_box_array = Array.new
	inclusion_box_array = Array.new
	tmp_array = Array.new

	inclusion_box_array = doc_incl_box_list.get(0, "end")
	exclusion_box_array = code_excl_box_list.get(0, "end")

	box_indices_array = doc_box_list.curselection

	box_indices_array.each { |i|

		box = doc_box_list.get(i)

		case
			when (inclusion_box_array.include?(box)): process_errors(1, box, "E")
			when (exclusion_box_array.include?(box)): process_errors(2, box, "E")
			else inclusion_box_array.push(box)
		end
	}

	inclusion_box_array.each { |box|
		box.scan(/(box)(\d+)(_)(\d+)/) { tmp_array.push([$2.to_i,$4.to_i]) }
	}

	tmp_array.sort!
	doc_incl_box_list.delete(0, "end")

	tmp_array.each_index { |i|
		doc_incl_box_list.insert("end", "box#{tmp_array[i][0]}_#{tmp_array[i][1]}")
	}
}

code_box_sel_mode_button.bind("Button-1") {

	# Determine the current selectmode configuration, it will be either "extended", one or
	# more contiguous items selectable at a time, or "multiple", where more than one 
	# non-contiguous items selectable at a time. Switch configuration between the two: 
	# "browse" or "multiple".
	case code_box_list.cget("selectmode")

		when "extended"

			code_box_list.configure("selectmode", "multiple")
			code_box_sel_mode_button.configure("text", "Select Multiple")

		when "multiple"

			# User can select (highlight) multiple list elements at once.
			code_box_list.configure("selectmode", "extended")
			code_box_sel_mode_button.configure("text", "Select Single")

			# Clear all selected elements, then select (highlight) the last element selected,
			# which is the currently 'active' element - the last one clicked on.
			code_box_list.selection_clear(0, "end")
			code_box_list.selection_set(code_box_list.index("active"))
	end
}

doc_box_sel_mode_button.bind("Button-1") {

	# Determine the current selectmode configuration, it will be either "extended", one or
	# more contiguous items selectable at a time, or "multiple", where more than one 
	# non-contiguous items selectable at a time. Switch configuration between the two: 
	# "browse" or "multiple".
	case doc_box_list.cget("selectmode")

		when "extended"

			doc_box_list.configure("selectmode", "multiple")
			doc_box_sel_mode_button.configure("text", "Select Multiple")

		when "multiple"

			# User can select (highlight) multiple list elements at once.
			doc_box_list.configure("selectmode", "extended")
			doc_box_sel_mode_button.configure("text", "Select Single")

			# Clear all selected elements, then select (highlight) the last element selected,
			# which is the currently 'active' element - the last one clicked on.
			doc_box_list.selection_clear(0, "end")
			doc_box_list.selection_set(doc_box_list.index("active"))
	end
}

clear_input_button.bind("Button-1") {

	# Clear all objects associated with the input code..
	in_text.delete("1.0", "end")
	code_box_list.delete(0, "end")
	code_excl_box_list.delete(0, "end")

	# Set the focus to the input text window, and put the cursor at the first char on line one.
	in_text.focus
	in_text.mark_set('insert', '1.0')
}

s_blank_button.bind("Button-1") {

	# Perform single mode blanking and if the OFF/ON toggle buttons are "ON" we will also delete the corresponding list(s).
	blank_button(from_SciTE, "single", out_text, in_text, code_excl_box_list, doc_incl_box_list, custom_bar_text.value, multi_boxes_group_num.value.to_i)

	if (from_SciTE)

		exit(0)
	else
		code_box_list.delete(0, code_box_list.index('end')) if (code_box_clear_mode_button.cget("text") == "ON")
		code_excl_box_list.delete(0, code_excl_box_list.index('end')) if (code_excl_clear_mode_button.cget("text") == "ON")
		doc_box_list.delete(0, doc_box_list.index('end')) if (doc_box_clear_mode_button.cget("text") == "ON")
		doc_incl_box_list.delete(0, doc_incl_box_list.index('end')) if (doc_incl_clear_mode_button.cget("text") == "ON")
	end
}

m_blank_button.bind("Button-1") {

	# Perform multiple mode blanking and if the OFF/ON toggle buttons are "ON" we will also delete the corresponding list(s).
	blank_button(from_SciTE, "multiple", out_text, in_text, code_excl_box_list, doc_incl_box_list, custom_bar_text.value, multi_boxes_group_num.value.to_i)

	if (from_SciTE)

		exit(0)
	else
		code_box_list.delete(0, code_box_list.index('end')) if (code_box_clear_mode_button.cget("text") == "ON")
		code_excl_box_list.delete(0, code_excl_box_list.index('end')) if (code_excl_clear_mode_button.cget("text") == "ON")
		multi_boxes_group_num.value = "" if (multi_clear_mode_button.cget("text") == "ON")
	end
}

c_blank_button.bind("Button-1") {

	# Perform custom mode blanking and if the OFF/ON toggle buttons are "ON" we will also delete the corresponding vars.
	blank_button(from_SciTE, "custom", out_text, in_text, code_excl_box_list, doc_incl_box_list, custom_bar_text.value, multi_boxes_group_num.value.to_i)

	if (from_SciTE)

		exit(0)
	else
		custom_bar_text.value = "" if (cust_clear_mode_button.cget("text") == "ON")
	end
}

r_blank_button.bind("Button-1") {
	reverse_blank_button(from_SciTE, out_text, in_text)
	exit(0) if (from_SciTE)
}

get_boxes_button.bind("Button-1") { process_input_code(code_box_list, in_text) }

code_box_clear_button.bind("Button-1") { code_box_list.delete(0, code_box_list.index('end')) }

#
# This will delete selected boxes from the code_excl_box_list.
#
code_excl_clear_sel_button.bind("Button-1") {

	tmp_array = Array.new

	indices = code_excl_box_list.curselection

	if (indices.size > 0)

		0.upto(code_excl_box_list.size - 1) do |i|

			tmp_array.push(code_excl_box_list.get(i)) unless (indices.include?(i)) 
		end

		code_excl_box_list.delete(0, "end")

		tmp_array.each { |box| code_excl_box_list.insert("end", box) }
	end
}

#
# This will delete selected boxes from the doc_incl_box_list.
#
doc_incl_clear_sel_button.bind("Button-1") {

	tmp_array = Array.new

	indices = doc_incl_box_list.curselection

	if (indices.size > 0)

		0.upto(doc_incl_box_list.size - 1) do |i|

			tmp_array.push(doc_incl_box_list.get(i)) unless (indices.include?(i)) 
		end

		doc_incl_box_list.delete(0, "end")

		tmp_array.each { |box| doc_incl_box_list.insert("end", box) }
	end
}

code_excl_clear_button.bind("Button-1") { code_excl_box_list.delete(0, code_excl_box_list.index('end')) }

code_box_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the code_box_list when either the single or multi blank button is selected.
	#
	case code_box_clear_mode_button.cget("text")

		when "OFF" then code_box_clear_mode_button.configure("text", "ON")
		when "ON" then code_box_clear_mode_button.configure("text", "OFF")
	end
	}

code_excl_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the code_excl_box_list when either the single or multi blank button is selected.
	#
	case code_excl_clear_mode_button.cget("text")

		when "OFF" then code_excl_clear_mode_button.configure("text", "ON")
		when "ON" then code_excl_clear_mode_button.configure("text", "OFF")
	end
	}

doc_box_clear_button.bind("Button-1") { doc_box_list.delete(0, doc_box_list.index('end')) }

doc_incl_clear_button.bind("Button-1") { doc_incl_box_list.delete(0, doc_incl_box_list.index('end')) }

doc_box_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the doc_box_list when either the single or multi blank button is selected.
	#
	case doc_box_clear_mode_button.cget("text")

		when "OFF" then doc_box_clear_mode_button.configure("text", "ON")
		when "ON" then doc_box_clear_mode_button.configure("text", "OFF")
	end
	}

doc_incl_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the doc_incl_box_list when either the single or multi blank button is selected.
	#
	case doc_incl_clear_mode_button.cget("text")

		when "OFF" then doc_incl_clear_mode_button.configure("text", "ON")
		when "ON" then doc_incl_clear_mode_button.configure("text", "OFF")
	end
	}

cust_clear_button.bind("Button-1") { custom_bar_text.value = "" }

cust_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the custom_bar_text when either the single or multi blank button is selected.
	#
	case cust_clear_mode_button.cget("text")

		when "OFF" then cust_clear_mode_button.configure("text", "ON")
		when "ON" then cust_clear_mode_button.configure("text", "OFF")
	end
	}

multi_clear_button.bind("Button-1") { multi_boxes_group_num.value = "" }

multi_clear_mode_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the custom_bar_text when either the single or multi blank button is selected.
	#
	case multi_clear_mode_button.cget("text")

		when "OFF" then multi_clear_mode_button.configure("text", "ON")
		when "ON" then multi_clear_mode_button.configure("text", "OFF")
	end
	}

delete_box_list_button.bind("Button-1") {

	# Clears all box lists.
	code_box_list.delete(0, code_box_list.index('end'))
	code_excl_box_list.delete(0, code_excl_box_list.index('end'))
	doc_box_list.delete(0, doc_box_list.index('end'))
	doc_incl_box_list.delete(0, doc_incl_box_list.index('end'))
}

#
# ###############################################################
#
# This handles the tool startup from the SciTE tools menu. The Ruby Code file is read to obtain the list of boxes
# in the document, and any selected code [highlighted] is automatically read and inserted into the code input area
# and the associated code box list is populated.
# 
if (ARGV[0])

	#
	# We need to ensure this application was started from a document's Ruby code file i.e. has a file name
	# in AAAAxxxx.rbw where A is any alpha character and x is any digit.
	#
	if (File.basename(ARGV[0]) =~ /([a-zA-Z]{4})\d{4}\.rbw/)

		#
		# Now flag that we started this tool from SciTE so we will automatically shut down when the
		# blank/reverse blank button has been clicked on.
		#
		from_SciTE = true

		#
		# This is the whole thing - path and file name.
		#
		cur_file = ARGV[0]

		process_doc_file(cur_file, doc_box_list_label, doc_box_list)

		if (ARGV[1])

			# We have highlighted [selected] Ruby Code.
			in_text.insert("end", ARGV[1])

			process_input_code(code_box_list, in_text)
		end
	else
		process_errors(16, nil, "E")
	end
end

#
# This calls exit_app( ) when the user clicks on the 'X' button to close the application window so we
# can prevent the application output from writing/overwriting [when selected text] the SciTE file.
#
# This is because this application can run as a stand alone app in which case we do not automatically write to the
# SciTE file OR it can run from the SciTE Tools menu as a command in which case the app will close upon blanking/reverse
# blanking or could close when the user Exits the app from the menu or exits from the 'X' button.
#
TkRoot.new.protocol('WM_DELETE_WINDOW', proc{exit_app(from_SciTE)})

Tk.mainloop()
